;;;; -*- Mode: Lisp -*-

;;;; format-formatting.lisp
;;;; Portably (mostly) formatting (IEEE) floats in Common Lisp.
;;;;
;;;; This code is put in the Public Domain, but please do mention its
;;;; origin:
;;;;
;;;; Marco Antoniotti
;;;; DCB, DISCo,
;;;; Universita` degli Studi di Milano-Bicocca, Milan, Italy
;;;; 2023-11-12.
;;;;
;;;; The code uses second-hand information from Wikipedia:
;;;;
;;;; en.wikipedia.org/wiki/Floating-point_arithmetic
;;;; en.wikipedia.org/wiki/Single-precision_floating-point_format
;;;; en.wikipedia.org/wiki/Double-precision_floating-point_format
;;;;
;;;; Some of the constants' value and hardwired numbers you see in
;;;; this code are taken from the above references.


;;; Packaging all-in-one.

(defpackage "IT.UNIMIB.DISCO.MA.FLOAT-FORMATTING" (:use "CL")
  (:nicknames "FLOAT-FMT")
  (:documentation "The FLOAT-FORMATTING Package.

This package offers minimal facilities to see the binary representation
of a Common Lisp FLOAT.

The main function exported is FORMAT-FLOAT-BINARY.
")

  (:export
   "FLOAT-BITS"
   "FLOAT-EXP-BIAS"
   "FORMAT-FLOAT-BINARY"
   )
  )


(in-package "FLOAT-FMT")

;;; float-layout
;;; A placeholder for future developments.

(deftype float-layout ()
  '(member :ieee :ieee-floating-point))


;;; Useful constants.
;;; Feel free to fix the defaults.

(defconstant +short-float-bits+ #+lispworks 32 #-lispworks 32)
(defconstant +single-float-bits+ #+lispworks 32 #-lispworks 32) 
(defconstant +double-float-bits+ #+lispworks 64 #-lispworks 64)
(defconstant +long-float-bits+ #+lispworks 64 #-lispworks 64)


(defconstant +short-float-exp-bias+ #+lispworks 127 #-lispworks 127)
(defconstant +single-float-exp-bias+ #+lispworks 127 #-lispworks 127) 
(defconstant +double-float-exp-bias+ #+lispworks 1023 #-lispworks 1023)
(defconstant +long-float-exp-bias+ #+lispworks 1023 #-lispworks 1023)


;;; float-bits --

(defun float-bits (&optional (f 1.0f0) (layout :ieee))
  "Returns how many bits the float F occupies in memory."
  (declare (ignore layout))
  (etypecase f
    (short-float  +short-float-bits+)
    (single-float +single-float-bits+)
    (double-float +double-float-bits+)
    (long-float   +long-float-bits+)
    ))


;;; float-exp-bias --

(defun float-exp-bias (&optional (f 1.0f0) (layout :ieee))
  "Returns the bias in the exponent field in a IEEE representation of float F."
  (declare (ignore layout))
  (etypecase f
    (short-float  +short-float-exp-bias+)
    (single-float +single-float-exp-bias+)
    (double-float +double-float-exp-bias+)
    (long-float   +long-float-exp-bias+)
    ))


(declaim (ftype (function (&optional float float-layout)
                          (integer 1 256)) ; Ok.  Too many assumptions.
                float-bits)
         (ftype (function (&optional float float-layout)
                          (integer 1 16383)) ; Ok.  Too many assumptions.
                float-exp-bias)
         (inline float-bits float-exp-bias))


;;; format-float-binary
;;; Writes the binary representation of a FLOAT F.
;;;
;;; Notes:
;;; (INTEGER-DECODE-FLOAT F) returns as first value an integer I which
;;; is then encoded as
;;;
;;;   (- I (expt 2 (1- (float-digits F))))
;;;
;;; in the significand portion of the floating point binary
;;; representation.
;;;
;;; This means that the second value E returned by I-D-F is not what
;;; ends up encoded in the exponent field (after accounting for BIAS);
;;; what gets encoded is actually
;;;
;;;   (+ BIAS (float-digits F) E)
;;;

(defun format-float-binary (f &key
                              ((:stream out) *standard-output*)
                              (base 2)
                              (sep #\Space)
                              (layout :ieee)
                              )
  "Writes the binary representation of a FLOAT F.

The FLOAT F binary representation is written out to the stream
:STREAM.  The representation fields - sign, exponent and significand -
are written out using BASE (default 2) and SEP (default #\Space)
separating them.  If :STREAM is NIL, a string containing the
representation is returned.

Notes:

While the code is potentially portable, it eventually assumes standard
IEEE field widths and the presence of a 'hidden' bit.  This may be
'fixed' in a future version; for the time being the LAYOUT arguments
other than :IEEE or :IEEE-FLOATING-POINT cause a warning to be issued
and the function to return.
"

  (declare (type float f)
           (type (or null t stream) out)
           (type (integer 2 16) base)
           (type character sep)
           )

  (unless (typep layout 'float-layout)
    (warn "FFB: cannot handle layout ~S (yet)." layout)
    (if (null out)
        (return-from format-float-binary
          (format nil "#<Float ~S is in ~S format>" f layout))
        (return-from format-float-binary nil)))

  (let* ((precision (float-precision f))
         (digits    (float-digits f))
         (bits      (float-bits f))
         (bias      (float-exp-bias f))
         (exp-field (- bits
                       1 ; Sign.
                       (1- digits) ; Bits sans "hidden bit".
                       ))
         (format-cntl ; Ok: let's just be reasonable...
          (case base
            (2 "~C~B~C~V,'*B~C~V,'*B~C")
            (8 "~C~O~C~V,'*O~C~V,'*O~C")
            (10 "~C~D~C~V,'*D~C~V,'*D~C")
            (16 "~C~X~C~V,'*X~C~V,'*X~C")
            (otherwise "~C~B~C~V,'*B~C~V,'*B~C") ; Same as '2'.
            ))
         (signif-sans-hidden-bit 0)
         (is-denormalized (and (not (zerop f)) (/= precision digits)))
         )
    (declare (type fixnum precision digits bits bias exp-field)
             (type fixnum signif-sans-hidden-bit) ; FTTB.
             (type boolean is-denormalized)
             (type string format-cntl)
             )

    (when is-denormalized
      (warn "FFB: ~F is a denormalized float with precision ~D and digits ~D."
            f
            precision
            digits))

    (multiple-value-bind (signif exp sign)
        (integer-decode-float f)
      (declare (type integer signif exp) ; Too wide.
               (type (integer -1 1) sign)) ; Too wide.

      #+nil
      (format *debug-io*
              "FFB: significand integer length: ~D~%FFB: |~VB|~%"
              (integer-length signif)
              (integer-length signif)
              signif
              )
      
      (when (= (integer-length signif) digits) ; Could it be > ?
        ;; Remove the 'hidden bit'.  Love these kludges!
        (setf signif-sans-hidden-bit
              (dpb 0 (byte 1 (1- digits)) signif)))

      (format out format-cntl
              sep
              (if (minusp sign) 1 0)
              sep
              exp-field (+ bias (1- digits) exp) ; NEED the digits
                                                 ; here to account for
                                                 ; the "hidden bit".
              sep
              (1- digits) signif-sans-hidden-bit ; digits - "hidden bit".
              sep
              ))
    ))

;;;; end of file -- float-formatting.lisp
